import { Inject } from "@nestjs/common";
import { CommandHandler, ICommandHandler } from "@nestjs/cqrs";
import { ConflictException } from "@src/libs/exceptions";
import { Err, Ok, Result } from "oxide.ts";
import { UserRepositoryPort } from "../database/user.repository.port";
import { UserEntity } from "../domain/user.entity";
import { UserAlreadyExistsError } from "../domain/user.error";
import { UserResponseDto } from "../dtos/user.response.dto";
import { USER_REPOSITORY } from "../user.di-tokens";
import { CreateUserCommand } from "./create-user.command";

@CommandHandler(CreateUserCommand)
export class CreateUserService implements ICommandHandler {
  constructor(
    @Inject(USER_REPOSITORY)
    protected readonly userRepo: UserRepositoryPort
  ) {}

  async execute(
    command: CreateUserCommand
  ): Promise<Result<UserResponseDto, UserAlreadyExistsError>> {
    const user = UserEntity.create({
      email: command.email,
      password: command.password,
      fullname: command.fullname,
      phone_number: command.phone_number,
      role_id: command.role_id
    });

    try {
      await this.userRepo.transaction(async () => this.userRepo.insert(user));

      const userResponse = new UserResponseDto(user);
      userResponse.email = command.email;
      userResponse.fullname = command.fullname;
      userResponse.phone_number = command.phone_number;
      userResponse.role_id = command.role_id;

      return Ok(userResponse)
    } catch (error: any) {
      if(error instanceof ConflictException) {
        return Err(new UserAlreadyExistsError(error));
      }
      throw error;
    }
  }
}