import { Injectable, Logger } from '@nestjs/common';
import { EventEmitter2 } from '@nestjs/event-emitter';
import { SqlRepositoryBase } from '@src/libs/db/sql-repository.base';
import { InjectPool } from 'nestjs-slonik';
import { DatabasePool, sql } from 'slonik';
import { z } from 'zod';
import { UserEntity } from '../domain/user.entity';
import { UserMapper } from '../user.mapper';
import { UserRepositoryPort } from './user.repository.port';

export const userSchema = z.object({
  id: z.string().uuid(),
  email: z.string().email(),
  phone_number: z.string().min(10).max(15),
  password: z.string(),
  fullname: z.string(),
  avatar:  z.string().optional(),
  latitude:  z.string().optional(),
  longitude:  z.string().optional(),
  auth_token:  z.string().optional(),
  verify_token:  z.string().optional(),
  fcm_token:  z.string().optional(),
  is_verified:  z.string().optional(),
  is_merchant:  z.string().optional(),
  role_id: z.number(),
  created_at: z.preprocess((val: any) => new Date(val), z.date()),
  updated_at: z.preprocess((val: any) => new Date(val), z.date()),
});


export type UserModel = z.TypeOf<typeof userSchema>;

/**
 *  Repository is used for retrieving/saving domain entities
 * */
@Injectable()
export class UserRepository 
  extends SqlRepositoryBase<UserEntity, UserModel> 
  implements UserRepositoryPort 
{
  protected tableName = 'users';
  
  protected schema = userSchema;

  constructor(
    @InjectPool()
    pool: DatabasePool,
    mapper: UserMapper,
    eventEmitter: EventEmitter2
  ) {
    super(pool, mapper, eventEmitter, new Logger(UserRepository.name));
  } 

  async findOneByEmail(email: string): Promise<UserEntity> {
    const user = await this.pool.one(
      sql.type(userSchema)`SELECT * FROM "users" WHERE email = ${email}`
    )

    return this.mapper.toDomain(user);
  }
  
}